﻿//using NAudio.Wave;
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Runtime.InteropServices;
using System.Text;
using System.Threading.Tasks;
using System.IO;
using static SDL2.SDL;

namespace SharpNes.Audio
{
    public class ApuAudioProvider : IDisposable// : WaveProvider32
    {
        static ApuAudioProvider()
        {
            if (SDL_Init(SDL_INIT_AUDIO | SDL_INIT_TIMER) != 0)
            {
                Debug.WriteLine("Could not initialize SDL - %s\n", SDL_GetError());
            }
        }
        SDL_AudioSpec spec;

        public ApuAudioProvider()
        {
            cyclicBuffer = new float[4096];
            readIndex = writeIndex = 0;
            Enabled = true;

            spec.freq = 22050;
            spec.format = AUDIO_F32SYS;
            spec.channels = 1;
            spec.silence = 0;
            spec.samples = 1024;
            spec.callback = Read;

            //open audio devcie
            if (SDL_OpenAudio(ref spec, out var audioSpec) != 0)
            {
                Debug.WriteLine("Failed to open audio device, %s\n", SDL_GetError());
            }
        }
        bool isEnabled = true;
        public bool Enabled
        {
            get { return isEnabled; }
            set
            {
                isEnabled = value;
                if (value)
                {
                    SDL_PauseAudio(0);
                }
                else
                {
                    SDL_PauseAudio(1);
                }
            }
        }

        public unsafe void Read(IntPtr userdata, IntPtr buffer, int len)
        {
            lock (queueLock)
            {
                if (!isEnabled || size == 0)
                {
                    for (int i = 0; i < len; i++)
                    {
                        ((byte*)buffer)[i] = 0;
                    }
                    //return;
                }

                //var data = (byte*)buffer;
                var sampleCount = Math.Min(len / 4, size);
                //List<byte> scr = new List<byte>();
                for (int n = 0; n < sampleCount; n++)
                {
                    var f = cyclicBuffer[readIndex++];
                    ((float*)buffer)[n] = f;
                    readIndex %= cyclicBuffer.Length;
                    --size;
                }
                //SDL_MixAudio(buffer, scr.ToArray(), (uint)scr.Count, SDL_MIX_MAXVOLUME);
            }
        }

        public void Queue(float[] sampleValues)
        {
            lock (queueLock)
            {
                for (int index = 0; index < sampleValues.Length; index++)
                {
                    if (size >= cyclicBuffer.Length)
                        return;

                    cyclicBuffer[writeIndex] = sampleValues[index];
                    ++writeIndex;
                    writeIndex %= cyclicBuffer.Length;
                    ++size;
                }
            }
        }

        public void Dispose()
        {
            //file.Close();
            SDL_CloseAudio();
        }

        private float[] cyclicBuffer = new float[8192];
        private int readIndex;
        private int writeIndex;
        private int size;
        private object queueLock = new object();
    }


    [StructLayout(LayoutKind.Explicit, Pack = 2)]
    public class WaveBuffer
    {
        /// <summary>
        /// Number of Bytes
        /// </summary>
        [FieldOffset(0)]
        public int numberOfBytes;

        [FieldOffset(8)]
        private byte[] byteBuffer;

        [FieldOffset(8)]
        private float[] floatBuffer;

        [FieldOffset(8)]
        private short[] shortBuffer;

        [FieldOffset(8)]
        private int[] intBuffer;

        /// <summary>
        /// Gets the byte buffer.
        /// </summary>
        /// <value>The byte buffer.</value>
        public byte[] ByteBuffer => byteBuffer;

        /// <summary>
        /// Gets the float buffer.
        /// </summary>
        /// <value>The float buffer.</value>
        public float[] FloatBuffer => floatBuffer;

        /// <summary>
        /// Gets the short buffer.
        /// </summary>
        /// <value>The short buffer.</value>
        public short[] ShortBuffer => shortBuffer;

        /// <summary>
        /// Gets the int buffer.
        /// </summary>
        /// <value>The int buffer.</value>
        public int[] IntBuffer => intBuffer;

        /// <summary>
        /// Gets the max size in bytes of the byte buffer..
        /// </summary>
        /// <value>Maximum number of bytes in the buffer.</value>
        public int MaxSize => byteBuffer.Length;

        /// <summary>
        /// Gets or sets the byte buffer count.
        /// </summary>
        /// <value>The byte buffer count.</value>
        public int ByteBufferCount
        {
            get
            {
                return numberOfBytes;
            }
            set
            {
                numberOfBytes = CheckValidityCount("ByteBufferCount", value, 1);
            }
        }

        /// <summary>
        /// Gets or sets the float buffer count.
        /// </summary>
        /// <value>The float buffer count.</value>
        public int FloatBufferCount
        {
            get
            {
                return numberOfBytes / 4;
            }
            set
            {
                numberOfBytes = CheckValidityCount("FloatBufferCount", value, 4);
            }
        }

        /// <summary>
        /// Gets or sets the short buffer count.
        /// </summary>
        /// <value>The short buffer count.</value>
        public int ShortBufferCount
        {
            get
            {
                return numberOfBytes / 2;
            }
            set
            {
                numberOfBytes = CheckValidityCount("ShortBufferCount", value, 2);
            }
        }

        /// <summary>
        /// Gets or sets the int buffer count.
        /// </summary>
        /// <value>The int buffer count.</value>
        public int IntBufferCount
        {
            get
            {
                return numberOfBytes / 4;
            }
            set
            {
                numberOfBytes = CheckValidityCount("IntBufferCount", value, 4);
            }
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="T:NAudio.Wave.WaveBuffer" /> class.
        /// </summary>
        /// <param name="sizeToAllocateInBytes">The number of bytes. The size of the final buffer will be aligned on 4 Bytes (upper bound)</param>
        public WaveBuffer(int sizeToAllocateInBytes)
        {
            int num = sizeToAllocateInBytes % 4;
            sizeToAllocateInBytes = ((num == 0) ? sizeToAllocateInBytes : (sizeToAllocateInBytes + 4 - num));
            byteBuffer = new byte[sizeToAllocateInBytes];
            numberOfBytes = 0;
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="T:NAudio.Wave.WaveBuffer" /> class binded to a specific byte buffer.
        /// </summary>
        /// <param name="bufferToBoundTo">A byte buffer to bound the WaveBuffer to.</param>
        public WaveBuffer(byte[] bufferToBoundTo)
        {
            BindTo(bufferToBoundTo);
        }

        /// <summary>
        /// Binds this WaveBuffer instance to a specific byte buffer.
        /// </summary>
        /// <param name="bufferToBoundTo">A byte buffer to bound the WaveBuffer to.</param>
        public void BindTo(byte[] bufferToBoundTo)
        {
            byteBuffer = bufferToBoundTo;
            numberOfBytes = 0;
        }

        /// <summary>
        /// Performs an implicit conversion from <see cref="T:NAudio.Wave.WaveBuffer" /> to <see cref="T:System.Byte" />.
        /// </summary>
        /// <param name="waveBuffer">The wave buffer.</param>
        /// <returns>The result of the conversion.</returns>
        public static implicit operator byte[](WaveBuffer waveBuffer)
        {
            return waveBuffer.byteBuffer;
        }

        /// <summary>
        /// Performs an implicit conversion from <see cref="T:NAudio.Wave.WaveBuffer" /> to <see cref="T:System.Single" />.
        /// </summary>
        /// <param name="waveBuffer">The wave buffer.</param>
        /// <returns>The result of the conversion.</returns>
        public static implicit operator float[](WaveBuffer waveBuffer)
        {
            return waveBuffer.floatBuffer;
        }

        /// <summary>
        /// Performs an implicit conversion from <see cref="T:NAudio.Wave.WaveBuffer" /> to <see cref="T:System.Int32" />.
        /// </summary>
        /// <param name="waveBuffer">The wave buffer.</param>
        /// <returns>The result of the conversion.</returns>
        public static implicit operator int[](WaveBuffer waveBuffer)
        {
            return waveBuffer.intBuffer;
        }

        /// <summary>
        /// Performs an implicit conversion from <see cref="T:NAudio.Wave.WaveBuffer" /> to <see cref="T:System.Int16" />.
        /// </summary>
        /// <param name="waveBuffer">The wave buffer.</param>
        /// <returns>The result of the conversion.</returns>
        public static implicit operator short[](WaveBuffer waveBuffer)
        {
            return waveBuffer.shortBuffer;
        }

        /// <summary>
        /// Clears the associated buffer.
        /// </summary>
        public void Clear()
        {
            Array.Clear(byteBuffer, 0, byteBuffer.Length);
        }

        /// <summary>
        /// Copy this WaveBuffer to a destination buffer up to ByteBufferCount bytes.
        /// </summary>
        public void Copy(Array destinationArray)
        {
            Array.Copy(byteBuffer, destinationArray, numberOfBytes);
        }

        /// <summary>
        /// Checks the validity of the count parameters.
        /// </summary>
        /// <param name="argName">Name of the arg.</param>
        /// <param name="value">The value.</param>
        /// <param name="sizeOfValue">The size of value.</param>
        private int CheckValidityCount(string argName, int value, int sizeOfValue)
        {
            int num = value * sizeOfValue;
            if (num % 4 != 0)
            {
                throw new ArgumentOutOfRangeException(argName, $"{argName} cannot set a count ({num}) that is not 4 bytes aligned ");
            }
            if (value >= 0 && value <= byteBuffer.Length / sizeOfValue)
            {
                return num;
            }
            throw new ArgumentOutOfRangeException(argName, $"{argName} cannot set a count that exceed max count {byteBuffer.Length / sizeOfValue}");
        }
    }
}
